/*
** Module   : DirtyShareFactory.java
** Abstract : Factory for DirtyShareContext and DirtyShareManager instances
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 ECF 20080403 @38630   Created initial version. Factory for DirtyShareContext and 
**                           DirtyShareManager instances.
** 002 SVL 20080806 @39309   Added sessions mapping and functions for deregistration.  
** 003 CA  20140513          Added a weight for the context-local var, to ensure predetermined 
**                           order during context reset.
*/

package com.goldencode.p2j.persist.dirty;

import java.util.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.remote.*;
import com.goldencode.p2j.security.*;

/**
 * Factory for DirtyShareContext and DirtyShareManager instances.
 */
public final class DirtyShareFactory
{
   /** Cache of {@link DefaultDirtyShareManager} instances by database */
   private static final Map<Database, DirtyShareManager> cache =
      new HashMap<Database, DirtyShareManager>();

   /** Map of DirtyShareManager to Session instances. */
   private static final Map<DirtyShareManager, Session> sessions =
      new HashMap<DirtyShareManager, Session>();
   
   /** Context local map of Databases to DirtyShareContext instances */
   private static final ContextLocal<Map<Database, DirtyShareContext>> local =
      new ContextLocal<Map<Database, DirtyShareContext>>()
   {
      public WeightFactor getWeight()
      {
         return WeightFactor.LEVEL_7;
      };
      
      protected Map<Database, DirtyShareContext> initialValue()
      {
         return (new HashMap<Database, DirtyShareContext>());
      }
      
      protected void cleanup(Map<Database, DirtyShareContext> map)
      {
         for (DirtyShareContext ctx : map.values())
         {
            ctx.cleanup();
         }
      }
   };
   
   /**
    * Get the dirty share context for the given, primary database.  This
    * method will first check a local cache for the context.  If not present
    * in the cache, the context is created and cached for future retrievals.
    * For the temporary table database, a do-nothing implementation is
    * returned.
    * 
    * @param   database
    *          Primary database which is the key for the dirty share context
    *          in the local cache.
    * 
    * @return  Dirty share context associated with the given database.
    */
   public static DirtyShareContext getContextInstance(Database database)
   {
      Map<Database, DirtyShareContext> map = local.get();
      DirtyShareContext ctx = map.get(database);
      if (ctx == null)
      {
         if (DatabaseManager.TEMP_TABLE_DB.equals(database))
         {
            ctx = new NopDirtyShareContext();
         }
         else
         {
            DirtyShareManager dsm = getManagerInstance(database);
            ctx = new DefaultDirtyShareContext(dsm);
         }
         
         map.put(database, ctx);
      }
      
      return ctx;
   }

   /**
    * Removes the dirty share share context for the given database from the
    * local cache, if any.
    *
    * @param database
    *        Primary database which is the key for the dirty share context
    *        in the local cache.
    */
   public static void unregisterContextInstance(Database database)
   {
      Map<Database, DirtyShareContext> map = local.get();
      DirtyShareContext ctx = map.get(database);

      if (ctx != null)
      {
         map.remove(database);
      }
   }
   
   /**
    * Get the dirty share manager for the given, primary database.  This
    * method will first check a shared cache for the manager.  If not present
    * in the cache, the manager is created and cached for future retrievals.
    * For local databases, an instance of {@link DefaultDirtyShareManager}
    * is returned.  For remote databases, an instance of {@link
    * com.goldencode.p2j.persist.remote.RemoteDirtyShareManager} is returned.
    * 
    * @param   database
    *          Primary database which is the key for the dirty share manager
    *          in the shared cache.
    * 
    * @return  Dirty share manager associated with the given database.
    */
   public static DirtyShareManager getManagerInstance(Database database)
   {
      if (DatabaseManager.TEMP_TABLE_DB.equals(database))
      {
         throw new IllegalArgumentException(
            "Cannot create DirtyShareManager for temporary table database)");
      }
      
      synchronized (cache)
      {
         DirtyShareManager dsm = cache.get(database);

         if (dsm == null)
         {
            if (database.isLocal())
            {
               dsm = new DefaultDirtyShareManager(database);
            }
            else
            {
               // Retrieve network session.
               Session session = ConnectionManager.getSession(database);

               // Get dirty share manager multiplexer proxy.
               DirtyShareMultiplexer multiplexer = (DirtyShareMultiplexer)
                  RemoteObject.obtainNetworkInstance(
                     DirtyShareMultiplexer.class, session);

               // Instantiate the remote dirty share manager.
               RemoteDirtyShareManager rdsm =
                  new RemoteDirtyShareManager(multiplexer);
               rdsm.setDatabase(database);
               dsm = rdsm;

               sessions.put(dsm, session);
            }
            cache.put(database, dsm);
         }
         
         return dsm;
      }
   }

   /**
    * Determines whether the given session for the given database is used
    * by the dirty share manager for this database.
    *
    * @param database
    *        Primary database which is the key for the dirty share manager
    *        in the shared cache.
    * @param session
    *        Candidate for checking whether it is used by the shared maanger.
    *
    * @return <code>true</code> of the session is used by the dirty share
    *         manager for this database, <code>false</code> otherwise.
    */
   public static boolean isUsedByManager(Database database, Session session)
   {
      synchronized(cache)
      {
         DirtyShareManager dsm = cache.get(database);
         return dsm != null && session == sessions.get(dsm);
      }
   }

   /**
    * Removes the dirty share manager for the given database from the shared
    * cache, if any. Also terminates and removes from the cache session used
    * by this manager.
    *
    * @param database
    *        Primary database which is the key for the dirty share manager
    *        in the shared cache.
    */
   public static void unregisterManagerForDatabase(Database database)
   {
      Session session = null;
      synchronized(cache)
      {
         DirtyShareManager dsm = cache.get(database);
         if (dsm != null)
         {
            cache.remove(database);

            session = sessions.remove(dsm);
         }
      }

      if (session != null)
      {
         session.terminate();
      }
   }
}
