/*
** Module   : SecurityContext.java
** Abstract : Class that manages security contexts for P2J server threads
**            representing remote applications and clients.
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 NVS 20050307   @20216 Created. This class manages security contexts
** 002 NVS 20050427   @20911 Enhanced to provide unlimited storage for any
**                           kind of tokens (objects) that can be associa-
**                           ted with this context. The need is to get rid
**                           of all sorts of per thread state information
**                           and replace it with per context one. This
**                           overcomes the problem of arbitrary chosen
**                           threads from the thread pool to serve eleme-
**                           ntary requests which have to share some state
**                           for the lifespan of the session.
** 003 NVS 20050504   @21067 unassign() method is now responsible for the
**                           security context cleanup, which is due when 
**                           the use counter reaches 0. Added a private
**                           method called cleanup() where this processing
**                           takes place.
** 004 NVS 20050506   @21116 Security contexts are now bound to generation
**                           of security cache they are built on. This
**                           change help in fixing the security cache 
**                           refresh bug. Added private member to keep the
**                           reference to the cache and getCache() method.
** 005 ECF 20060120   @24004 Modified {add|remove|has|get}Token methods.
**                           Replaced String token name with an Object
**                           key.
** 006 ECF 20060123   @24035 Added read-only session ID. The unique ID is
**                           set by the SecurityManager at SecurityContext
**                           construction. One session ID is reserved for
**                           the server/system context.
** 007 NVS 20070412   @32965 Reworked security context cleanup. cleanup()
**                           method is made package private. It isn't
**                           called from unassign() now. Instead, unassign
**                           returns a boolean indicating the state of
**                           the context.
** 008 ECF 20070708   @34418 Optimized getToken() and removeToken().
**                           Removed unnecessary checks preceding actual
**                           retrieval/removal.
** 009 NVS 20080513   @38273 Fixed javadoc for unassign().
** 010 ECF 20081006   @40040 Added safety logic and logging to cleanup().
**                           Cleanable's cleanup() calls are now enclosed
**                           in a try-catch block, with logging on error.
** 011 CA  20091119   @44431 Added SessionToken implementation - an unique
**                           object is kept for each security context, which
**                           will hold the session ID and the subject.
** 012 GES 20111021          Added the transferAllTokens() method.
** 013 CA  20120727          The cleanup() method should check if the payload
**                           is an instance of Cleanable - testing if one of
**                           its directly implemented interfaces is Cleanable
**                           poses a risk of not identifying a payload which
**                           does not directly implement Cleanable, but its
**                           superclass (or someone else up the hierarchy)
**                           does.
**                           Also, the tokenMap is cleared at the beginning of
**                           cleanup(), to ensure no other thread can start
**                           the cleaning process again.
** 014 CA  20130628          Added possibility to reset a context (needed by appservers running in
**                           State-reset operating mode).
** 015 CA  20140413          Fix for H014 - ensure the context-local vars are cleaned up in a
**                           predetermined order, when the security context is reset.
**                           Context cleanup uses same weights as context-reset, except all vars 
**                           are cleaned, unconditionally.
*/ 

package com.goldencode.p2j.security;

import java.util.*;

/**
 * Security context is a collection of control data that associates subjects
 * with Access Control Lists in the P2J directory. 
 */
class SecurityContext
{
   /** Reserved session ID for system threads */
   static final int SYSTEM_SESSION = 0;

   /** flags the initial security context */
   private boolean isInitial = false;

   /** index of process account */
   private int procAcc = -1;

   /** index of user account */
   private int userAcc = -1;

   /** indices of group accounts */
   private int[] grpsAcc = null;

   /** combined rights check list */
   private int[] accList = null;

   /** use count for this context */
   private int use = 0;

   /** caches access decisions made under this context */
   private Map decisionCache[] = null;

   /** saved procId */
   private int procId = -1;

   /** saved userId */
   private int userId = -1;

   /** saved groupIds */
   private int[] groupId = null;

   /** a map of arbitrary tokens that are associated with the context */
   private Map tokenMap = null;

   /** security cache this context stack is bound to */
   private SecurityCache cache = null;
   
   /** unique session token for this security context */
   private SessionToken sessionToken = null;

   /** Lock to avoid the cleanup to take place simultaneusly. */
   private final Object cleanupLock = new Object();
   
   /**
    * Package private constructor used to create a user or system session
    * context.
    *
    * @param  cache
    *         security cache to bind this context to
    * @param  proc
    *         index of the process account or -1
    * @param  user
    *         index of the user account or -1
    * @param  groups
    *         array of indices of the group accounts or null
    * @param  nres
    *         number of registered resources
    * @param  session
    *         Unique session ID for this context
    * @param  subject
    *         Subject authenticated in this context. This can only be 
    *         <code>null</code> when the <code>session</code> is set to the
    *         {@link #SYSTEM_SESSION}.
    * 
    * @throws NullPointerException
    *         if the <code>subject</code> is null and this is not a system
    *         session.
    */
   SecurityContext(SecurityCache cache, int proc, int user, int[] groups, 
                   int nres, int session, String subject)
   {
      if (session != SYSTEM_SESSION && subject == null)
      {
         throw new NullPointerException("Creating a security context with " +
         		"a null session token is not allowed !");
      }

      this.cache = cache;

      int len = groups != null ? groups.length : 0;

      procAcc = proc;
      userAcc = user;
      grpsAcc = groups;

      sessionToken = new SessionToken(session, subject);

      accList = new int[len + 1];
      if (user != -1)
         accList[0] = user;
      else
         accList[0] = proc;
      for (int i = 0; i < len; i ++)
         accList[i + 1] = groups[i];

      decisionCache = new Map[nres];
      for (int i = 0; i < nres; i ++)
         decisionCache[i] = new HashMap();

      procId = proc;
      userId = user;
      groupId = groups;

      tokenMap = new HashMap();
   }

   /**
    * Package private constructor used to create a system session context.
    *
    * @param cache
    *        security cache to bind this context to
    * @param proc
    *        index of the process account or -1
    * @param user
    *        index of the user account or -1
    * @param groups
    *        array of indices of the group accounts or null
    * @param nres
    *        number of registered resources
    * @param initial
    *        flag for initial security context
    */
   SecurityContext(SecurityCache cache, int proc, int user, int[] groups, 
                   int nres, boolean initial)
   {
      this(cache, proc, user, groups, nres, SYSTEM_SESSION, null);
      isInitial = initial;
   }

   /**
    * Marks this security context as assigned by incrementing its use count.
    */
   synchronized void assign()
   {
      use ++;
   }

   /**
    * Marks this security context as unassigned by decrementing its use count.
    *
    * @return  <code>true</code> if the use count has reached 0, otherwise
    *          <code>false</code>
    */
   synchronized boolean unassign()
   {
      use --;
      if (use == 0)
      {
         return true;
      }
      
      return false;
   }

   /**
    * Gets the array of accounts associated with this context.
    *
    * @return array of indices into accounts
    */
   int[] getCheckList()
   {
      return accList;
   }

   /**
    * Gets the array of original subject IDs of this context.
    *
    * @return array of indices into accounts
    */
   int[] getIdList()
   {
      int groups = groupId == null ? 0 : groupId.length;
      int[] list = new int[2 + groups];

      list[0] = procId;
      list[1] = userId;
      for (int i = 0; i < groups; i ++)
         list[2 + i] = groupId[i];

      return list;
   }

   /**
    * Checks whether this context is initial.
    *
    * @return <code>true</code> if this context is initial
    */
   boolean isInitial()
   {
      return isInitial;
   }

   /**
    * Gets process account index.
    *
    * @return process account index
    */
   int getProcessOrdinal()
   {
      return procAcc;
   }

   /**
    * Gets user account index.
    *
    * @return user account index
    */
   int getUserOrdinal()
   {
      return userAcc;
   }
   
   /**
    * Gets session ID.
    *
    * @return session ID.
    */
   int getSessionId()
   {
      return sessionToken.getSessionID();
   }

   /**
    * Gets session token.
    *
    * @return session token.
    */
   SessionToken getSessionToken()
   {
      return sessionToken;
   }

   /**
    * Gets the access decision cache.
    *
    * @return array of maps containing access decisions
    */
   Map[] getDecisionCache()
   {
      return decisionCache;
   }

   /**
    * Gets the security cache this context is bound to.
    *
    * @return <code>SecurityCache</code>
    */
   SecurityCache getCache()
   {
      return cache;
   }

   /**
    * Adds a token to the context map.
    *
    * @param key
    *        token key to be used as a key in the map
    *
    * @param token
    *        an arbitrary object to be kept in the entry
    *
    * @return <code>true</code> if the token is added successfully
    */
   synchronized boolean addToken(Object key, Object token)
   {
      if (tokenMap.containsKey(key))
         return false;

      tokenMap.put(key, token);
      return true;
   }

   /**
    * Remove a token from the context map.
    *
    * @param key
    *        token key to be used as a key in the map
    *
    * @return <code>true</code> if the token was removed successfully
    */
   synchronized boolean removeToken(Object key)
   {
      return (tokenMap.remove(key) != null);
   }

   /**
    * Checks whether the specified token is in the context map.
    *
    * @param key
    *        token key to be used as a key in the map
    *
    * @return <code>true</code> if the token is found
    */
   synchronized boolean hasToken(Object key)
   {
      return tokenMap.containsKey(key);
   }

   /**
    * Gets the token from the context map.
    *
    * @param key
    *        token key to be used as a key in the map
    *
    * @return the value of the token as a generic object or <code>null</code>
    */
   synchronized Object getToken(Object key)
   {
      return tokenMap.get(key);
   }

   /**
    * Copies all tokens to the given context and clears the token map of the
    * current instance.
    */
   synchronized void transferAllTokens(SecurityContext other)
   {
      other.tokenMap.putAll(this.tokenMap);
      this.tokenMap.clear();
   }

   /**
    * Reset this security context.
    */
   void reset()
   {
      cleanupWorker(true);
   }
   
   /**
    * Finalizes this security context.
    */
   void cleanup()
   {
      // avoid cleanup to be called multiple times
      synchronized (cleanupLock)
      {
         cleanupWorker(false);
      }
   }
   
   /**
    * Actual worker which fully cleans or just resets the context-local data.
    * 
    * @param    reset
    *           When {@code true}, this flag resets the security context; else, full cleanup is
    *           performed.
    */
   private void cleanupWorker(boolean reset)
   {
      // using a cloned map because we might clear the tokenMap
      Map<Object, Object> clone = new HashMap<>();
      synchronized (this)
      {
         clone.putAll(tokenMap);
      }

      // for debugging purposes only
      final boolean debug = false;
      if (debug)
      {
         ContextLocal.setDebug(true);
      }
      
      List<Object> allKeys = new ArrayList<>();
      List<WeightedToken> weightedKeys = new ArrayList<>();
      List<Object> remainder = new ArrayList<>();

      // collect all weighted tokens and sort them
      Iterator<Object> iter = clone.keySet().iterator();
      while (iter.hasNext())
      {
         Object key = iter.next();
         if (key instanceof WeightedToken)
         {
            weightedKeys.add((WeightedToken) key);
         }
         else
         {
            remainder.add(key);
         }
      }
      
      Collections.sort(weightedKeys, new Comparator<WeightedToken>()
      {
         @Override
         public int compare(WeightedToken o1, WeightedToken o2)
         {
            return o1.getWeight().compareTo(o2.getWeight());
         }
      });
      
      // add all un-weighted keys first and the weighted keys last
      allKeys.addAll(remainder);
      allKeys.addAll(weightedKeys);
      
      iter = allKeys.iterator();

      // get every token from the context
      while (iter.hasNext())
      {
         Object key = iter.next();
         Object payload = clone.get(key);

         if (payload == null || 
             (reset && key instanceof ContextLocal && !((ContextLocal) key).isResetAllowed()))
         {
            continue;
         }
         
         // check whether this token implements Cleanable interface
         if (Cleanable.class.isAssignableFrom(payload.getClass()))
         {
            try
            {
               ((Cleanable) payload).cleanup();
            }
            catch (Exception exc)
            {
               SecurityManager secMgr = SecurityManager.getInstance();
               if (secMgr != null)
               {
                  secMgr.debug(SecurityManager.ERROR,
                               "Error cleaning up security context",
                               exc);
               }
            }
         }
         
         // remove it from the clone map
         clone.remove(key);
      }
      
      // for debugging purposes only
      if (debug)
      {
         ContextLocal.dumpDependencies();
      }

      // add the surviving entries back to the tokenMap
      synchronized (this)
      {
         // check if tokenMap contains other data beside what is left in the clone
         List<String> failedTokens = new ArrayList<>();
         for (Object key : tokenMap.keySet())
         {
            if (!clone.containsKey(key))
            {
               failedTokens.add(key.getClass().getName());
            }
         }
         
         if (!failedTokens.isEmpty())
         {
            SecurityManager secMgr = SecurityManager.getInstance();
            if (secMgr != null)
            {
               final String msg = "SecurityContext.cleanupWorker did not complete properly - " +
                                  "the following tokens are still in use: %s";
               secMgr.debug(SecurityManager.ERROR, String.format(msg, failedTokens.toString()));
            }
         }
         
         // even if reset was not successful, ensure only the expected tokens will survive the
         // reset
         tokenMap.clear();
         tokenMap.putAll(clone);
      }
   }
}
